import collections
from typing import Callable, Dict, KeysView, Any
from gatenlp.offsetmapper import OffsetMapper, OFFSET_TYPE_JAVA, OFFSET_TYPE_PYTHON
from gatenlp.annotation_set import AnnotationSet
from gatenlp.annotation import Annotation
from gatenlp.changelog import ChangeLog
from .feature_bearer import FeatureBearer
import logging

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class _AnnotationSetsDict(collections.defaultdict):
    """
    A dict name to annotationset which creates and stores an empty annotation
    set on the fly when it is requested.
    """
    def __init__(self, changelog: ChangeLog = None, owner_doc: "Document" = None):
        super().__init__()
        self.changelog = changelog
        self.owner_doc = owner_doc

    def __missing__(self, key: str):
        annset = AnnotationSet(name=key, changelog=self.changelog, owner_doc=self.owner_doc)
        self[key] = annset
        return annset


class Document(FeatureBearer):
    """
    Represent a GATE document. This is different from the original Java GATE representation in several ways:

    * the text is not mutable and can only be set at creation time, so there is no "edit" method
    * as a feature bearer, all the methods to set, get and manipulate features are part of this class, there is
      no separate "FeatureMap" to store them
    * does not support listener callbacks
    * there is no separate abstraction for "content", the only content possible is text which is a unicode string
      that can be acessed with the "text()" method
    * Spans of text can be directly accessed using doc[from:to]
    * features are not stored in a separate feature map object, but are directly set on the document, e.g.
      doc.set_feature("x",y) or doc.get_feature("x", defaultvalue)
    * Features may only have string keys and values which can be json-serialised
    * Annotation offsets by default are number of Unicde code points, this is different from Java where the offsets
      are UTF-16 Unicode code units
    * Offsets of all annotations can be changed from/to Java (from python index of unicode codepoint to Java index
      of UTF-16 code unit and back)
    * No part of the document has to be present, not even the text (this allows saving just the annotations separately
      from the text)
    * Once the text has been set, it is immutable (no support to edit text and change annotation offsets accordingly)

    :param text: the text of the document. The text can be None to indicate that no initial text should be set. Once
      the text has been set for a document, it is immutable and cannot be changed.
    :param features: the initial document features to set, a sequence of key/value tuples
    :param changelog: a ChangeLog instance to use to log changes.
    """
    def __init__(self, text: str, features=None, changelog: ChangeLog = None):
        super().__init__(features)
        self.gatenlp_type = "Document"
        self.changelog = changelog
        self.annotation_sets = _AnnotationSetsDict(self.changelog, owner_doc=self)
        self._text = text
        self.offset_type = OFFSET_TYPE_PYTHON
        self.document_type = "simple"
        self.repr_version = '1.0'

    def _ensure_type_python(self) -> None:
        if self.offset_type != OFFSET_TYPE_PYTHON:
            raise Exception("Document cannot be used if it is not type PYTHON, use to_type(OFFSET_TYPE_PYTHON) first")

    def _fixup_annotations(self, method: Callable) -> None:
        annset_names = self.annotation_sets.keys()
        for annset_name in annset_names:
            annset = self.annotation_sets[annset_name]
            for ann in annset:
                ann.start = method(ann.start)
                ann.end = method(ann.end)

    def to_type(self, offsettype: str) -> None:
        """
        Convert all the offsets of all the annotations in this document to the
        required type, either OFFSET_TYPE_JAVA or OFFSET_TYPE_PYTHON. If the offsets
        are already of that type, this does nothing.

        :param offsettype: either OFFSET_TYPE_JAVA or OFFSET_TYPE_PYTHON
        :return:
        """
        if offsettype == self.offset_type:
            return
        if offsettype == OFFSET_TYPE_JAVA and self.offset_type == OFFSET_TYPE_PYTHON:
            # convert from currently python to java
            om1 = OffsetMapper(self._text)
            self._fixup_annotations(om1.convert_to_java)
            self.offset_type = OFFSET_TYPE_JAVA
        elif offsettype == OFFSET_TYPE_PYTHON and self.offset_type == OFFSET_TYPE_JAVA:
            # convert from currently java to python
            om1 = OffsetMapper(self._text)
            self._fixup_annotations(om1.convert_to_python)
            self.offset_type = OFFSET_TYPE_PYTHON
        else:
            raise Exception("Odd offset type")

    def set_changelog(self, chlog: ChangeLog) -> ChangeLog:
        """
        Make the document use the given changelog to record all changes
        from this moment on.

        :param chlog: the new changelog to use or None to not use any
        :return: the changelog used previously or None
        """
        oldchlog = self.changelog
        self.changelog = chlog
        # make sure all the inner data structures use that chlog as well:
        self.annotation_sets.changelog = chlog
        for k, v in self.annotation_sets.items():
            v.changelog = chlog
        return oldchlog

    @property
    def text(self) -> str:
        """
        Get the text of the document. For a partial document, the text may be None.

        :return: the text of the document
        """
        self._ensure_type_python()
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        """
        Set the text of the document. This is only possible as long as it has not been set
        yet, after that, the text is immutable.

        :param value: the text for the document
        :return:
        """
        if self._text is None:
            self._text = value
        else:
            raise NotImplementedError("Text cannot be modified")

    def size(self) -> int:
        """
        Return the size of the document text.
        Note: this will convert the type of the document to python!

        :return: size of the document (length of the text)
        """
        self._ensure_type_python()
        return int(len(self.text))

    def _log_feature_change(self, command: str, feature: str = None, value=None) -> None:
        if self.changelog is None:
            return
        ch = {"command": command}
        if command == "feature:set":
            ch["feature"] = feature
            ch["value"] = value
        self.changelog.append(ch)

    def __len__(self) -> int:
        """
        Return the length of the text.
        Note: this will convert the type of the document to python!

        :return: the length of the document text
        """
        self._ensure_type_python()
        return len(self._text)

    def __getitem__(self, span) -> str:
        """
        Get the text for the given span.

        :param span: a single number, an offset range of the form from:to or an annotation.
        If annotation, uses the annotation's offset span.
        :return: the text of the span
        """
        self._ensure_type_python()
        if isinstance(span, Annotation):
            return self.text[span.start:span.end]
        return self.text[span]

    def get_annotations(self, name: str = "") -> AnnotationSet:
        """
        Get the named annotation set, if name is not given or the empty string, the default annotation set.
        If the annotation set does not already exist, it is created.

        :param name: the annotation set name, the empty string is used for the "default annotation set".
        :return: the specified annotation set.
        """
        self._ensure_type_python()
        return self.annotation_sets[name]

    def get_annotation_set_names(self) -> KeysView[str]:
        """
        Return the set of known annotation set names.

        :return: annotation set names
        """
        self._ensure_type_python()
        return self.annotation_sets.keys()
    
    def remove_annotation_set(self, name: str):
        """
        Completely remove the annotation set.
        :param name: name of the annotation set to remove
        :return:
        """
        del self.annotation_sets[name]
        if self.changelog:
            self.changelog.append({
                "command": "annotations:remove",
                "set": self.name})

    def __repr__(self) -> str:
        """
        String representation of the document, showing all content.

        :return: string representation
        """
        return "Document({},features={},anns={})".format(self.text, self.features, self.annotation_sets)

    def _json_repr(self, **kwargs) -> Dict:
        """
        Return a a simple map representation of this document for JSON to serialize.

        :return: something JSON can serialize
        """
        offset_type = self.offset_type
        if "offset_type" in kwargs and kwargs["offset_type"] != offset_type:
            om = OffsetMapper(self._text)
            kwargs["offset_mapper"] = om
            offset_type = kwargs["offset_type"]
        return {
            "text": self._text,
            "features": self.features,
            # turn our special class into an ordinary map
            "annotation_sets": {name: annset._json_repr(**kwargs) for name, annset in self.annotation_sets.items()},
            "offset_type": offset_type,
            "document_type": self.document_type,
            "repr_version": self.repr_version,
            "gatenlp_type": self.gatenlp_type
        }

    @staticmethod
    def _from_json_map(jsonmap: Dict[str, Any], **kwargs) -> "Document":
        """
        Construct a document instance from the JSON map representation we get.

        :param jsonmap: the map representation of a document used for JSON
        :param kwargs: any kwargs passed through from the load/loads method
        :return: a document instance
        """
        doc = Document(jsonmap.get("text"), features=jsonmap.get("features"))
        doc.annotation_sets = _AnnotationSetsDict()
        for k, v in jsonmap.get("annotation_sets").items():
            # print("Adding set {} of type {}".format(k, type(v)), file=sys.stderr)
            doc.annotation_sets[k] = v
        offset_type = jsonmap.get("offset_type")
        doc.offset_type = offset_type
        if offset_type == OFFSET_TYPE_JAVA:
            doc.to_type(OFFSET_TYPE_PYTHON)
        if "with_changelog" in kwargs:
            chlog = ChangeLog()
            doc.set_changelog(chlog)
        doc.repr_version = jsonmap.get("repr_version")
        doc.document_type = jsonmap.get("document_type")
        return doc



"""
A "mixin" class to use for classes which should have features. This turns the class into something
that can be used much like a dict in addition to original methods of the class. The inheriting class
must implement _log_feature_change(command, feature=None, value=None) and must have an attribute
changelog.
"""

from typing import List, Union, Dict, Set, KeysView


class FeatureBearer:

    def __init__(self, initialfeatures=None):
        """
        Initialise any features, if necessary.
        :param initialfeatures: an iterable containing tuples of initial feature key/value pairs
        :return:
        """
        if initialfeatures is not None:
            self.features = dict(initialfeatures)
        else:
            self.features = None
        self.changelog = None  # this must be set by the inheriting class!

    def _log_feature_change(self, command: str,
                            feature: Union[str, None] = None, value: Union[str, None] = None):
        """
        This should be overriden by the inheriting class!
        :param command: the command to log
        :param feature: the feature name involved. If the command is any that needds a feature name,
        the invoking method needs to make sure that the feature name is not None and also otherwise
        something that is allowed as a key.
        :param value: the value involved. None is a proper value if the feature is set. Otherwise, the value
        is always ignored.
        :return:
        """
        raise Exception("must be overridden by the inheriting class")

    def clear_features(self) -> None:
        """
        Remove all features.
        :return:
        """
        # if we do not have features, this is a NOOP
        if self.features is None:
            return
        self._log_feature_change("features:clear")
        # instead of emptying the dict, we remove it comepletely, maybe it wont be used anyway
        self.features = None

    def set_feature(self, key: str, value) -> None:
        """
        Set feature to the given value
        :param key: feature name
        :param value: value
        :return:
        """
        if self.features is None:
            self.features = dict()
        if key is None or not isinstance(key, str):
            raise Exception("A feature name must be a string, not {}".format(type(key)))
        self._log_feature_change("feature:set", feature=key, value=value)
        self.features[key] = value

    def del_feature(self, featurename: str) -> None:
        """
        Remove the feature with that name
        :param featurename: the feature to remove from the set
        :return:
        """
        if self.features is None:
            raise KeyError(featurename)
        self._log_feature_change("feature:remove", feature=featurename)
        del self.features[featurename]

    def get_feature(self, key: str, default=None):
        if self.features is None:
            return default
        return self.features.get(key, default)

    def has_feature(self, key: str) -> bool:
        if self.features is None:
            return False
        return key in self.features

    def feature_names(self) -> Union[Set, KeysView]:
        """
        Return an iterable with the feature names. This is NOT a view and does not update when the features change!
        :return:
        """
        if self.features is None:
            return set()
        else:
            return set(self.features.keys())

    def feature_values(self) -> List:
        """
        Return an iterable with the feature values. This is NOT a view and does not update when the features change!
        :return:
        """
        if self.features is None:
            return []
        else:
            return [set(self.features.values())]

    def features(self) -> Dict:
        """
        Return a shallow copy of the feature map. This is NOT a view and does not update when the features change!
        :return:
        """
        if self.features is None:
            return {}
        else:
            return self.features.copy()

    def update_features(self, *other, **kwargs):
        """
        Update the features from another map or an iterable of key value pairs or from keyword arguments
        :param other: another dictionary or an iterable of key,value pairs
        :param kwargs: used to update the features
        :return:
        """
        if self.features is None:
            self.features = {}
        if other:
            if hasattr(other, "keys"):
                for k in other:
                    self.set_feature(k, other[k])
            else:
                for k, v in other:
                    self.set_feature(k, v)
        if kwargs:
            for k in kwargs:
                self.set_feature(k, kwargs[k])

    def num_features(self) -> int:
        """
        Return the number of features. We do not use "len" for this, since the feature bearing object may
        have its own useful len implementation.
        :return: number of features
        """
        if self.features is None:
            return 0
        else:
            return len(self.features)



"""
An annotation is immutable, but the features it contains are mutable.
"""
import sys
from typing import List, Union, Dict, Set
from functools import total_ordering
from gatenlp.feature_bearer import FeatureBearer
from gatenlp.offsetmapper import OFFSET_TYPE_JAVA
from gatenlp.changelog import ChangeLog
from gatenlp._utils import support_annotation_or_set


@total_ordering
class Annotation(FeatureBearer):
    """
    NOTE: the current way how annotations are implemented tries to minimize the effort for storing the annotation
    in an annotation set.
    !! This means that equality and hashing outside of an annotation set will not work as expected: the identify
    of an annotation only depends on its annotation ID, relative to the set it is contained in. So if you want
    to store annotations from different sets you need to use (annotationset, id) as the unique identifier!
    """

    # We use slots to avoid the dict and save memory if we have a large number of annotations
    __slots__ = ('changelog', 'type', 'start', 'end', 'features', 'id', '_owner_setname')

    def __init__(self, start: int, end: int, annot_type: str, annot_id: int,
                 owner_set: "AnnotationSet" = None,
                 changelog: ChangeLog = None, features=None):
        """
        Create a new annotation instance. NOTE: this should almost never be done directly
        and instead the method annotation_set.add should be used!
        Once an annotation has been created, the start, end, type and id fields must not
        be changed!
        :param start: start offset of the annotation
        :param end: end offset of the annotation
        :param annot_type: annotation type
        :param annot_id: the id of the annotation
        :param owner_set: the containing annotation set
        :param changelog: a changelog, if changes to the features should get recorded
        :param features: an initial collection of features
        """
        super().__init__(features)
        self.gatenlp_type = "Annotation"
        # print("Creating Ann with changelog {} ".format(changelog), file=sys.stderr)
        self.changelog = changelog
        self.type = annot_type
        self.start = start
        self.end = end
        self.id = annot_id
        self._owner_set = owner_set

    # TODO: for now at least, make sure only simple JSON serialisable things are used! We do NOT
    # allow any user specific types in order to make sure what we create is interchangeable with GATE.
    # In addition we do NOT allow None features.
    # So a feature name always has to be a string (not None), the value has to be anything that is json
    # serialisable (except None keys for maps).
    # For performance reasons we check the feature name but not the value (maybe make checking optional
    # on by default but still optional?)
    def _log_feature_change(self, command: str, feature: str = None, value=None) -> None:
        if self.changelog is None:
            return
        ch = {
            "command": command,
            "set": self._owner_set.name,
            "id": self.id}
        if feature is not None:
            ch["feature"] = feature
        if value is not None:
            ch["value"] = value
        self.changelog.append(ch)

    def __eq__(self, other) -> bool:
        """
        Equality of annotations is only based on the annotation ID plus the owning set.
        :param other: the object to compare with
        :return: if the annotations are equal
        """
        if not isinstance(other, Annotation):
            return False
        if self._owner_set != other._owner_set:
            return False
        if self.id != other.id:
            return False
        else:
            return True

    def __hash__(self):
        """
        The hash depends on the annotation ID and the owning set.
        :return: hash
        """
        return hash((self.id, self._owner_set))

    def __lt__(self, other) -> bool:
        """
        Comparison for sorting: this sorts by increasing start offset, then increasing end offset, then increasing
        type name, then increasing annotation id.
        NOTE: for now the other object has to be an instance of Annotation, duck typing is not supported!
        :param other: another annotation
        :return:
        """
        if not isinstance(other, Annotation):
            raise Exception("Cannot compare to non-Annotation")
        if self.start < other.start:
            return True
        elif self.start > other.start:
            return False
        else:
            if self.end < other.end:
                return True
            elif self.end > other.end:
                return False
            else:
                if self.type < other.type:
                    return True
                elif self.type > other.type:
                    return False
                else:
                    if self.id < other.id:
                        return True
                    else:
                        return False

    def __repr__(self) -> str:
        """
        String representation of the annotation.
        :return: string representation
        """
        return "Annotation({},{},{},id={},features={})".format(self.start, self.end, self.type, self.id, self.features)

    def __len__(self) -> int:
        """
        The length of the annotation is the length of the offset span. Since the end offset is one after the last
        element, we return end-start-1
        :return:
        """
        return self.end - self.start - 1

    def inside(self, offset: int) -> bool:
        """
        Check if the given offset falls somewhere inside the span of this annotation.
        :param offset: the offset to check
        :return: True if the offset is inside the span of this annotation
        """
        return self.start <= offset < self.end

    @support_annotation_or_set
    def is_overlapping(self, start: int, end: int) -> bool:
        """
        Checks if this annotation is overlapping with the given span, annotation or
        annotation set.
        An annotation is overlapping with a span if the first or last character
        is inside that span.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: True if overlapping, False otherwise
        """
        return self.inside(start) or self.inside(end-1)

    @support_annotation_or_set
    def is_coextensive(self, start: int, end: int) -> bool:
        """
        Checks if this annotation is coextensive with the given span, annotation or
        annotation set, i.e. has exactly the same start and end offsets.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: True if coextensive, False otherwise
        """
        return self.start == start and self.end == end

    @support_annotation_or_set
    def is_within(self, start: int, end: int) -> bool:
        """
        Checks if this annotation is within the given span, annotation or
        annotation set, i.e. both the start and end offsets of this annotation
        are after the given start and before the given end.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: True if within, False otherwise
        """
        return start <= self.start and end >= self.end

    @support_annotation_or_set
    def is_covering(self, start: int, end: int) -> bool:
        """
        Checks if this annotation is covering the given span, annotation or
        annotation set, i.e. both the given start and end offsets
        are after the start of this annotation and before the end of this annotation.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: True if within, False otherwise
        """
        return self.start <= start and self.end >= end

    def _json_repr(self, **kwargs) -> Dict:
        if "offset_mapper" in kwargs:
            om = kwargs["offset_mapper"]
            to_type = kwargs["offset_type"]
            if to_type == OFFSET_TYPE_JAVA:
                start = om.convert_to_java(self.start)
                end = om.convert_to_java(self.end)
            else:
                start = om.convert_to_python(self.start)
                end = om.convert_to_python(self.end)
        else:
            start = self.start
            end = self.end
        return {
            "start": start,
            "end": end,
            "type": self.type,
            "id": self.id,
            "features": self.features,
            "gatenlp_type": self.gatenlp_type
        }

    @staticmethod
    def _from_json_map(jsonmap, **kwargs) -> "Annotation":
        ann = Annotation(jsonmap.get("start"), jsonmap.get("end"), jsonmap.get("type"), jsonmap.get("id"),
                         features=jsonmap.get("features"))
        return ann

    def __setattr__(self, key, value):
        """
        Prevent start, stop, type and annotation id from getting overridden, once they have been
        set.
        :param key: attribute to set
        :param value: value to set attribute to
        :return:
        """
        if key == "start" or key == "end" or key == "type" or key == "id":
            if self.__dict__.get(key, None) is None:
                super().__setattr__(key, value)
            else:
                raise Exception("Annotation attributes cannot get changed after being set")
        else:
            super().__setattr__(key, value)



from typing import List, Callable, Dict
from gatenlp.offsetmapper import OffsetMapper, OFFSET_TYPE_JAVA, OFFSET_TYPE_PYTHON


class ChangeLog:
    def __init__(self):
        self.gatenlp_type = "ChangeLog"
        self.changes = []
        self.offset_type = OFFSET_TYPE_PYTHON

    def append(self, element: Dict):
        assert isinstance(element, dict)
        self.changes.append(element)

    def __len__(self) -> int:
        return len(self.changes)

    def _fixup_changes(self, method: Callable) -> List[Dict]:
        """
        In-place modify the annotation offsets of the changes according to
        the given method.
        :param method: an object method method for converting offsets from or to python.
        :return: the modified changes, a reference to the modified changes list of the instance
        """
        for change in self.changes:
            if "start" in change:
                change["start"] = method(change["start"])
            if "end" in change:
                change["end"] = method(change["end"])
        return self.changes

    def fixup_changes(self, offset_mapper, offset_type):
        """
        In-place update the offsets of all annotations in this changelog to the desired
        offset type, if necessary. If the ChangeLog already has that offset type, this does nothing.
        :param offset_mapper: a prepared offset mapper to use
        :param offset_type: the desired offset type
        :return: a reference to the modified changes
        """
        if offset_type != self.offset_type:
            if offset_type == OFFSET_TYPE_JAVA:
                method = offset_mapper.convert_to_java
            elif offset_type == OFFSET_TYPE_PYTHON:
                method = offset_mapper.convert_to_python
            else:
                raise Exception("Not a proper offset type: {}".format(offset_type))
            return self._fixup_changes(method)

    def __repr__(self) -> str:
        return "ChangeLog([{}])".format(",".join([str(c) for c in self.changes]))

    def format_to(self, fp, prefix="") -> None:
        for c in self.changes:
            print(prefix, str(c), sep="", file=fp)

    def _json_repr(self, **kwargs) -> Dict:
        offset_type = self.offset_type
        changes = self.changes
        if "offset_type" in kwargs and kwargs["offset_type"] != offset_type:
            om = kwargs.get("offset_mapper")
            if om is None:
                raise Exception("Need to convert offsets, but no offset_mapper parameter given")
            offset_type = kwargs["offset_type"]
            if offset_type == OFFSET_TYPE_JAVA:
                changes = self._fixup_changes(om.convert_to_java)
            else:
                changes = self._fixup_changes(om.convert_to_python)
        return {
            "changes": changes,
            "offset_type": offset_type,
            "gatenlp_type": self.gatenlp_type
        }

    @staticmethod
    def _from_json_map(jsonmap, **kwargs) -> "ChangeLog":
        cl = ChangeLog()
        cl.changes = jsonmap.get("changes")
        cl.offset_type = jsonmap.get("offset_type")
        if cl.offset_type == OFFSET_TYPE_JAVA:
            # we need either an offset mapper or a document
            if "offset_mapper" in kwargs:
                om = kwargs.get("offset_mapper")
            elif "document" in kwargs:
                om = OffsetMapper(kwargs.get("document"))
            else:
                raise Exception("Loading a changelog with offset_type JAVA, need kwarg 'offset_mapper' or 'document'")
            cl._fixup_changes(om.convert_to_python)
        return cl



"""Convert between gatenlp and spacy documents and annotations."""

from gatenlp import interact, GateNlpPr, Document


def apply_nlp(nlp, gatenlpdoc, setname=""):
    """
    Run the spacy nlp pipeline on the gatenlp document and transfer the annotations.
    This modifies the gatenlp document in place.

    :param nlp: spacy pipeline
    :param gatenlpdoc: gatenlp document
    :param setname: set to use
    :return: 
    """
    spacydoc = nlp(gatenlpdoc.text)
    return spacy2gatenlp(spacydoc, gatenlpdoc=gatenlpdoc, setname=setname)


def spacy2gatenlp(spacydoc, gatenlpdoc=None, setname="", token_type="Token",
                  spacetoken_type="SpaceToken", sentence_type="Sentence",
                  nounchunk_type="NounChunk",
                  add_tokens=True, add_spacetokens=True,
                  add_ents=True, add_sents=True, add_nounchunks=True, add_dep=True):
    """
    Convert a spacy document to a gatenlp document. If a gatenlp document is already
    provided, add the annotations from the spacy document to it. In this case the
    original gatenlpdoc is used and gets modified.
    :param spacydoc: a spacy document
    :param gatenlpdoc: if None, a new gatenlp document is created otherwise this
    document is added to.
    :param setname: the annotation set name to which the annotations get added, empty string
    for the default annotation set.
    :param token_type: the annotation type to use for tokens
    :param spacetoken_type: the annotation type to use for space tokens
    :param sentence_type: the annotation type to use for sentence anntoations
    :param nounchunk_type: the annotation type to use for noun chunk annotations
    :param add_tokens: should annotations for tokens get added? If not, dependency parser
    info cannot be added either.
    :param add_spacetokens: should annotations for space tokens get added
    :param add_ents: should annotations for entities get added
    :param add_sents: should sentence annotations get added
    :param add_nounchunks: should noun chunk annotations get added
    :param add_dep: should dependency parser information get added
    :return: the new or modified
    """
    if gatenlpdoc is None:
        retdoc = Document(spacydoc.text)
    else:
        retdoc = gatenlpdoc
    toki2annid = {}
    annset = retdoc.get_annotations(setname)
    for tok in spacydoc:
        from_off = tok.idx
        to_off = tok.idx + len(tok)
        is_space = tok.is_space
        fm = {
                "_i": tok.i,
                "is_alpha": tok.is_alpha,
                "is_bracket": tok.is_bracket,
                "is_currency": tok.is_currency,
                "is_digit": tok.is_digit,
                "is_left_punct": tok.is_left_punct,
                "is_lower": tok.is_lower,
                "is_oov": tok.is_oov,
                "is_punct": tok.is_punct,
                "is_quote": tok.is_quote,
                "is_right_punct": tok.is_right_punct,
                "is_sent_start": tok.is_sent_start,
                "is_space": tok.is_space,
                "is_stop": tok.is_stop,
                "is_title": tok.is_title,
                "is_upper": tok.is_upper,
                "lang": tok.lang_,
                "lemma": tok.lemma_,
                "like_email": tok.like_email,
                "like_num": tok.like_num,
                "like_url": tok.like_url,
                "orth": tok.orth,
                "pos": tok.pos_,
                "prefix": tok.prefix_,
                "prob": tok.prob,
                "rank": tok.rank,
                "sentiment": tok.sentiment,
                "tag": tok.tag_,
                "shape": tok.shape_,
                "suffix": tok.suffix_,
        }
        if spacydoc.is_nered and add_ents:
            fm["ent_type"] = tok.ent_type_
        if spacydoc.is_parsed and add_dep:
            fm["dep"] = tok.dep_
        if tok.is_space:
            anntype = spacetoken_type
        else:
            anntype = token_type
        annid = annset.add(from_off, to_off, anntype, fm)
        toki2annid[tok.i] = annid
        # print("Added annotation with id: {} for token {}".format(annid, tok.i))
        ws = tok.whitespace_
        if len(ws) > 0:
            annset.add(to_off, to_off+len(ws), spacetoken_type, {"is_space": True})
    # if we have a dependency parse, now also add the parse edges
    if spacydoc.is_parsed and add_tokens and add_dep:
        for tok in spacydoc:
            ann = annset.get(toki2annid[tok.i])
            ann.set_feature("head", toki2annid[tok.head.i])
            ann.set_feature("left_edge", toki2annid[tok.left_edge.i])
            ann.set_feature("right_edge", toki2annid[tok.right_edge.i])
    if spacydoc.ents and add_ents:
        for ent in spacydoc.ents:
            annset.add(ent.start_char, ent.end_char, ent.label_, {"lemma": ent.lemma_})
    if spacydoc.sents and add_sents:
        for sent in spacydoc.sents:
            annset.add(sent.start_char, sent.end_char, sentence_type, {})
    if spacydoc.noun_chunks and add_nounchunks:
        for chunk in spacydoc.noun_chunks:
            annset.add(chunk.start_char, chunk.end_char, nounchunk_type, {})
    return retdoc




"""
GATE-specific (de)serialisation of documents. This is called "simplejson" to make it easy
to keep it apart from the default JSON de/serialiser (which is used but extended).
"""
import json
import gzip
from ..document import Document
from ..annotation import Annotation
from ..annotation_set import AnnotationSet
from ..changelog import ChangeLog


def get_object_encoder(**kwargs):
    """
    Returns a function for encoding our own objects. This simply checks if the object
    has the method "json_repr" and if yes, calls it with the kwargs we got.
    :return:
    """
    # Todo: check https://realpython.com/python-json/#encoding-and-decoding-custom-python-objects and similar
    # again for how to do this correctly. Instead of providing our own default method, maybe override
    # the JSONEncoder class: has the advantage that we can fallback to the default default method!
    def object_encoder(obj):
        if hasattr(obj, "_json_repr"):
            return obj._json_repr(**kwargs)
        else:
            # objtypename = obj.__class__.__name__
            raise TypeError("Cannot JSON-serialise {} of type {}".format(obj, type(obj)))
    return object_encoder


def get_object_hook(**kwargs):
    """
    Returns a method that will try to convert the passed map into one of our objects
    :param kwargs: the kwargs to use for converting back.
    :return: the object hook function
    """
    def object_hook(thedict):
        # NOTE: we need to explicitly see the type, duck typing could get mislead
        # by other objects that just happen to have similar fields!
        if not "gatenlp_type" in thedict:
            return thedict
        ourtype = thedict.get("gatenlp_type")
        if ourtype == "Document":
            return Document._from_json_map(thedict, **kwargs)
        elif ourtype == "Annotation":
            return Annotation._from_json_map(thedict, **kwargs)
        elif ourtype == "AnnotationSet":
            return AnnotationSet._from_json_map(thedict, **kwargs)
        elif ourtype == "ChangeLog":
            return ChangeLog._from_json_map(thedict, **kwargs)
        else:
            return thedict
    return object_hook


def load(fp, **kwargs):
    """
    Load gatenlp object from fp, a file-like object and return it.
    :param fp: a file-like object, as required by json.load
    :return: the gatenlp object
    """
    return json.load(fp, object_hook=get_object_hook(**kwargs))


def loads(str, **kwargs):
    """
    Create gatenlp object from JSON string and return it.

    :param str: JSON string
    :return: the gatenlp object
    """
    return json.loads(str, object_hook=get_object_hook(**kwargs))


def dump(fp, obj, indent=None, **kwargs):
    """
    Write the given gatenlp object to the file.
    :param fp: a file like object as required by json.dump
    :param obj: the object to save
    :param indent: passed on to jsom.dump
    :param kwargs:
    :return:
    """
    json.dump(fp, obj, indent=indent, default=get_object_encoder(**kwargs))


def dumps(obj, indent=None, **kwargs):
    """
    Create JSON string representing the given object.
    :param obj: the object
    :param indent: passed on to json.dumps
    :param kwargs:
    offset_type: if specified and OFFSET_TYPE_JAVA, convert the offsets to java offsets in the JSON
    offset_mapper: if specified, used for the offset mapping if an offset mapper cannot otherwise be found
    :return: JSON string
    """
    return json.dumps(obj, indent=indent, default=get_object_encoder(**kwargs))


def load_file(filename, **kwargs):
    """
    Shortcut for opening the file for reading and loading from the stream. If the filename ends with
    ".gz" the file is automatically uncompressed.
    :param filename: file to load
    :param kwargs:
    :return: the loaded object
    """
    if filename.endswith(".gz"):
        opener = gzip.open
        mode = "rt"
        encoding = "utf-8"
    else:
        opener = open
        mode = "rt"
        encoding = "utf-8"
    with opener(filename, mode, encoding=encoding) as fp:
        return load(fp, **kwargs)


def dump_file(obj, filename, indent=None, **kwargs):
    """
    Shortcut for opening the file for writing and dumping to the stream.
    If the file name ends with .gz, automatically compresses the output file.
    :param obj: the object to save
    :param filename: the file to write to
    :param indent: passed on to json.dump
    :param kwargs:
    :return:
    """
    if filename.endswith(".gz"):
        opener = gzip.open
        mode = "wt"
        encoding = "utf-8"
    else:
        opener = open
        mode = "wt"
        encoding = "utf-8"
    with opener(filename, mode, encoding=encoding) as fp:
        dump(obj, fp)



from typing import Any, List, Tuple, Union, Dict, Set, KeysView, Iterator, Generator
from collections import defaultdict
from gatenlp.annotation import Annotation
from gatenlp.exceptions import InvalidOffsetException
from gatenlp.changelog import ChangeLog
from gatenlp.impl import SortedIntvls
from gatenlp._utils import support_annotation_or_set


class AnnotationSet:
    def __init__(self, name: str = "", changelog: ChangeLog = None, owner_doc: "Document" = None):
        """
        Create a new annotation set.

        :param name: the name of the annotation set. This is only really needed if the changelog is used.
        :param changelog: if a changelog is used, then all changes to the set and its annotations are logged
        :param owner_doc: if this is set, the set and all sets created from it can be queried for the
        owning document and offsets get checked against the text of the owning document, if it has
        text.
        """
        # print("CREATING annotation set {} with changelog {} ".format(name, changelog), file=sys.stderr)
        self.gatenlp_type = "AnnotationSet"
        self.changelog = changelog
        self.name = name
        self.owner_doc = owner_doc

        # NOTE: the index is only created when we actually need it!
        # TODO: python 3.5
        # self._index_by_offset: SortedIntvls = None
        self._index_by_offset = None
        self._index_by_type = None
        # internally we represent the annotations as a map from annotation id (int) to Annotation
        self._annotations = {}
        self._is_immutable = False
        self._next_annid = 0

    def __setattr__(self, key, value):
        """
        Prevent immutable fields from getting overridden, once they have been
        set.
        :param key: attribute to set
        :param value: value to set attribute to
        :return:
        """
        if key == "gatenlp_type" or key == "name" or key == "owner_doc":
            if self.__dict__.get(key, None) is None:
                super().__setattr__(key, value)
            else:
                raise Exception("AnnotationSet attribute cannot get changed after being set")
        else:
            super().__setattr__(key, value)

    def restrict(self, restrict_to=None) -> "AnnotationSet":
        """
        Create an immutable copy of this set, optionally restricted to the given annotation ids.

        :param restrict_to: an iterable of annotation ids
        :return: an immutable annotation set with all the annotations of this set or restricted to the ids
          in restrict_to
        """
        annset = AnnotationSet(name="", owner_doc=self.owner_doc)
        annset._is_immutable = True
        annset._annotations = {annid: self._annotations[annid] for annid in restrict_to}
        annset._next_annid = self._next_annid
        return annset

    def _create_index_by_offset(self) -> None:
        """
        Generates the offset index, if it does not already exist.
        The offset index is an interval tree that stores the annotation ids for the offset interval of the annotation.
        """
        if self._index_by_offset is None:
            self._index_by_offset = SortedIntvls()
            for ann in self._annotations.values():
                self._index_by_offset.add(ann.start, ann.end, ann.id)

    def _create_index_by_type(self) -> None:
        """
        Generates the type index, if it does not already exist. The type index is a map from
        annotation type to a set of all annotation ids with that type.
        """
        if self._index_by_type is None:
            self._index_by_type = defaultdict(set)
            for ann in self._annotations.values():
                self._index_by_type[ann.type].add(ann.id)

    def _add_to_indices(self, annotation: Annotation) -> None:
        """
        If we have created the indices, add the annotation to them.

        :param annotation: the annotation to add to the indices.
        :return:
        """
        if self._index_by_type is not None:
            self._index_by_type[annotation.type].add(annotation.id)
        if self._index_by_offset is not None:
            self._index_by_offset.add(annotation.start, annotation.end, annotation.id)

    def _remove_from_indices(self, annotation: Annotation) -> None:
        """
        Remove an annotation from the indices.

        :param annotation: the annotation to remove.
        :return:
        """
        if self._index_by_offset is not None:
            self._index_by_offset.remove(annotation.start, annotation.end, annotation.id)
        if self._index_by_type is not None:
            self._index_by_type[annotation.type].remove(annotation.id)

    @staticmethod
    def _intvs2idlist(intvs) -> List[int]:
        """
        Convert an iterable of interval tuples (start, end, id) to a list of ids

        :param intvs: iterable of interval tuples
        :return: list of ids
        """
        return [i[2] for i in intvs]

    @staticmethod
    def _intvs2idset(intvs) -> Set[int]:
        """
        Convert an iterable of interval tuples (start, end, id) to a set of ids

        :param intvs: iterable of interval tuples
        :return: set of ids
        """
        ret = set()
        for i in intvs:
            ret.add(i[2])
        return ret

    def _restrict_intvs(self, intvs) -> "AnnotationSet":
        return self.restrict(AnnotationSet._intvs2idlist(intvs))

    def __len__(self) -> int:
        """
        Return number of annotations in the set.

        :return: number of annotations
        """
        return len(self._annotations)

    def size(self) -> int:
        """
        Return number of annotations in the set.

        :return: number of annotations
        """
        return len(self._annotations)

    def get_doc(self) -> Union["Document", None]:
        """
        Get the owning document, if known. If the owning document was not set, return None.

        :return: the document this annotation set belongs to or None if unknown.
        """
        return self.owner_doc

    @support_annotation_or_set
    def _check_offsets(self, start: int, end: int) -> None:
        """
        Checks the offsets for the given span/annotation against the document boundaries, if we know the owning
        document and if the owning document has text.
        """
        if self.owner_doc is None:
            return
        if self.owner_doc.text is None:
            return
        doc_size = self.owner_doc.size()

        if start < 0:
            raise InvalidOffsetException("Annotation starts before 0")
        if end < 0:
            raise InvalidOffsetException("Annotation ends before 0")
        if start > end:
            raise InvalidOffsetException("Annotation ends before it starts")
        if start > doc_size:
            raise InvalidOffsetException(
                "Annotation starts after document ends: start={}, docsize={}".format(start, doc_size))
        if end > doc_size:
            raise InvalidOffsetException(
                "Annotation ends after document ends: end={}, docsize={}".format(end, doc_size))

    def add(self, start: int, end: int, anntype: str, features, annid: int = None):
        """
        Add an annotation to the set. Once an annotation has been added, the start and end offsets,
        the type, and the annotation id are immutable.

        :param start: start offset
        :param end: end offset
        :param anntype: the annotation type
        :param features: a map, an iterable of tuples or an existing feature map. In any case, the features are used \
        to create a new feature map for this annotation.
        :param annid: the annotation id, if not specified the next free one for this set is used. NOTE: the id should\
        normally left unspecified and get assigned automatically.
        :return: the annotation id of the added annotation
        """
        if self._is_immutable:
            raise Exception("Cannot add an annotation to an immutable annotation set")
        self._check_offsets(start, end)
        if annid and annid in self._annotations:
            raise Exception("Cannot add annotation with id {}, already in set".format(annid))
        if annid is None:
            annid = self._next_annid
            self._next_annid = self._next_annid + 1
        ann = Annotation(start, end, anntype, annid, owner_set=self,
                         changelog=self.changelog, features=features)
        self._annotations[annid] = ann
        self._add_to_indices(ann)
        if self.changelog is not None:
            self.changelog.append({
                "command": "annotation:add",
                "set": self.name,
                "start": ann.start,
                "end": ann.end,
                "type": ann.type,
                "features": ann.features,
                "id": ann.id})
        return ann.id

    def add_ann(self, ann, annid: int = None):
        """
        Add a copy of the given ann to the annotation set, either with a new annotation id or
        with the one given.
        :param annid: the annotation id, if not specified the next free one for this set is used.
        NOTE: the id should normally left unspecified and get assigned automatically.
        :return: the annotation id of the added annotation
        """
        return self.add(ann.start, ann.end, ann.type, ann.features(), annid=annid)

    def remove(self, annotation: Union[int, Annotation]) -> None:
        """
        Remove the given annotation which is either the id or the annotation instance.

        :param annotation: either the id (int) or the annotation instance (Annotation)
        :return:
        """
        annid = None  # make pycharm happy
        if self._is_immutable:
            raise Exception("Cannot remove an annotation from an immutable annotation set")
        if isinstance(annotation, int):
            annid = annotation
            if annid not in self._annotations:
                raise Exception("Annotation with id {} not in annotation set, cannot remove".format(annid))
            annotation = self._annotations[annid]
        elif isinstance(annotation, Annotation):
            annid = annotation.id
            if annid not in self._annotations:
                raise Exception("Annotation with id {} does not belong to this set, cannot remove".format(annid))
        del self._annotations[annid]
        if self.changelog is not None:
            self.changelog.append({
                "command": "annotation:remove",
                "set": self.name,
                "id": annid})
        self._remove_from_indices(annotation)

    def clear(self) -> None:
        """
        Remove all annotations from the set.

        :return:
        """
        self._annotations.clear()
        self._index_by_offset = None
        self._index_by_type = None
        if self.changelog is not None:
            self.changelog.append({
                "command": "annotations:clear",
                "set": self.name})

    def __iter__(self) -> Iterator:
        """
        Iterator for going through all the annotations of the set.

        :return: a generator for the annotations in document order
        """
        # return iter(self._annotations.values())
        return self.in_document_order()

    def in_document_order(self, *annotations, from_offset: Union[int, None] = None,
                          to_offset: Union[None, int] = None,
                          reverse: bool = False, anntype: str = None) -> Generator:
        """
        Returns a generator for going through annotations in document order. If an iterator of annotations
        is given, then those annotations, optionally limited by the other parameters are returned in
        document order, otherwise, all annotations in the set are returned, otionally limited by the other
        parameters.

        :param annotations: either missing or exactly one parameter which must be an iterable of annotations
          from this annotation set.
        :param from_offset: the offset from where to start including annotations
        :param to_offset: the last offset to use as the starting offset of an annotation
        :param anntype: only annotations of this type
        :param reverse: process in reverse document order
        :return: generator for annotations in document order
        """
        if from_offset is not None:
            assert from_offset >= 0
        if to_offset is not None:
            assert to_offset >= 1
        if to_offset is not None and from_offset is not None:
            assert to_offset > from_offset
        if len(annotations) == 0:  # no annotations given, we use the ones in the set
            self._create_index_by_offset()
            for _start, _end, annid in self._index_by_offset.irange(minoff=from_offset, maxoff=to_offset, reverse=reverse):
                if anntype is not None and self._annotations[annid].type != anntype:
                    continue
                yield self._annotations[annid]
        elif len(annotations) == 1:
            for ann in sorted(annotations, reverse=reverse, key=lambda x: (x.start, x.end)):
                if anntype is not None and ann.type != anntype:
                    continue
                if from_offset is not None and ann.start < from_offset:
                    continue
                if to_offset is not None and ann.start > to_offset:
                    continue
                yield ann

    def get(self, annid: int, default=None) -> Union[Annotation, None]:
        """
        Gets the annotation with the given annotation id or returns the given default.

        :param annid: the annotation id of the annotation to retrieve.
        :param default: what to return if an annotation with the given id is not found.
        :return: the annotation or the default value.
        """
        return self._annotations.get(annid, default)

    def all_by_type(self, anntype: Union[str, None] = None) -> "AnnotationSet":
        """
        Gets annotations of the specified type. If the anntype is None, return all annotation in an immutable set.
        Creates the type index if necessary.

        :param anntype: if specified, the type of the annotations to return, of None, all annotations are selected.
        :return: an immutable annotation set with the matching annotations.
        """
        self._create_index_by_type()
        if anntype is None:
            annids = self._annotations.keys()
        else:
            annids = self._index_by_type.get(anntype, None)
            if annids is None:
                annids = []
        return self.restrict(annids)

    def type_names(self) -> KeysView[str]:
        """
        Gets the names of all types in this set. Creates the type index if necessary.

        :return: the set of known annotation type names.
        """
        self._create_index_by_type()
        return self._index_by_type.keys()

    @support_annotation_or_set
    def starting_at(self, start: int, ignored: Any = None) -> "AnnotationSet":
        """
        Gets all annotations starting at the given offset (empty if none) and returns them in an immutable
        annotation set.

        :param start: the offset where annotations should start
        :param ignored: dummy parameter to allow the use of annotations and annotation sets
        :return: annotation set of matching annotations
        """
        # NOTE: my assumption about how intervaltree works was wrong, so we need to filter what we get from the
        # point query
        self._create_index_by_offset()
        intvs = self._index_by_offset.starting_at(start)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def first_from(self, offset: int, ignored: Any = None) -> "AnnotationSet":
        """
        Gets all annotations at the first valid position at or after the given offset and returns them in an immutable
        annotation set.

        :param offset: The offset
        :param ignored: dummy parameter to allow the use of annotations and annotation sets
        :return: annotation set of matching annotations
        """
        self._create_index_by_offset()
        intvs = self._index_by_offset.starting_from(offset)
        # now select only those first ones which all have the same offset
        retids = set()
        startoff = None
        for intv in intvs:
            if startoff is None:
                startoff = intv[0]
                retids.add(intv[2])
            elif startoff == intv[0]:
                retids.add(intv[2])
            else:
                break
        return self.restrict(retids)

    @support_annotation_or_set
    def overlapping(self, start: int, end: int) -> "AnnotationSet":
        """
        Gets annotations overlapping with the given span. Instead of the start and end offsets,
        also accepts an annotation or annotation set.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: an immutable annotation set with the matching annotations
        """
        self._create_index_by_offset()
        intvs = self._index_by_offset.overlap(start, end)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def covering(self, start: int, end: int) -> "AnnotationSet":
        """
        Get the annotations which contain the given offset range (or annotation/annotation set)

        :param start: the start offset of the span
        :param end: the end offset of the span
        :return: an immutable annotation set with the matching annotations, if any
        """
        # This is not directly supported so we find the overlapping ones and then filter those where the
        # start and end fits
        # This may not be optimal
        self._create_index_by_offset()
        intvs = self._index_by_offset.covering(start, end)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def within(self, start: int, end: int) -> "AnnotationSet":
        """
        Gets annotations that fall completely within the given offset range

        :param start: start offset of the range
        :param end: end offset of the range
        :return: an immutable annotation set with the matching annotations
        """
        if start == end:
            intvs = []
        elif start > end:
            raise Exception("Invalid offset range: {},{}".format(start, end))
        else:
            self._create_index_by_offset()
            intvs = self._index_by_offset.within(start, end)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def starting_from(self, start: int, ignored: Any = None) -> "AnnotationSet":
        """
        Return the annotations that start at or after the given start offset.

        :param start: Start offset
        :param ignored: dummy parameter to allow the use of annotations and annotation sets
        :return: an immutable annotation set of the matching annotations
        """
        self._create_index_by_offset()
        intvs = self._index_by_offset.starting_from(start)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def starting_before(self, offset: int) -> "AnnotationSet":
        """
        Return the annotations that start before the given offset (or annotation). This also accepts an annotation
        or set.

        :param offset: offset before which the annotations should start
        :param ignored: dummy parameter to allow the use of annotations and annotation sets
        :return: an immutable annotation set of the matching annotations
        """
        self._create_index_by_offset()
        intvs = self._index_by_offset.starting_before(offset)
        return self._restrict_intvs(intvs)

    @support_annotation_or_set
    def coextensive(self, start: int, end: int) -> "AnnotationSet":
        """
        Return an immutable annotation set with all annotations that start and end at the given offsets.

        :param start: start offset of the span
        :param end: end offset of the span
        :return: annotation set with all annotations that have the same start and end offsets.
        """
        self._create_index_by_offset()
        intvs = self._index_by_offset.at(start, end)
        return self._restrict_intvs(intvs)

    def span(self) -> Tuple[int, int]:
        """
        Returns a tuple with the start and end offset the corresponds to the smallest start offset of any annotation
        and the largest end offset of any annotation.
        (Builds the offset index)

        :return: tuple of minimum start offset and maximum end offset
        """
        self._create_index_by_offset()
        return self._index_by_offset.min_start(), self._index_by_offset.max_end()

    def __contains__(self, annorannid: Union[int, Annotation]) -> bool:
        """
        Provides 'annotation in annotation_set' functionality

        :param annorannid: the annotation instance or annotation id to check
        :return: true if the annotation exists in the set, false otherwise
        """
        if isinstance(annorannid, Annotation):
            return annorannid.id in self._annotations
        return annorannid in self._annotations  # On the off chance someone passed an ID in directly

    contains = __contains__

    def __repr__(self) -> str:
        """
        String representation of the set.

        :return: string representation.
        """
        return "AnnotationSet({})".format(repr(list(self.in_document_order())))

    def _json_repr(self, **kwargs) -> Dict:
        return {
            "annotations": [ann._json_repr(**kwargs) for ann in self._annotations.values()],
            "next_annid": self._next_annid,
            "name": self.name,
            "gatenlp_type": self.gatenlp_type
        }

    @staticmethod
    def _from_json_map(jsonmap, **kwargs) -> "AnnotationSet":
        annset = AnnotationSet(name=jsonmap.get("name"))
        annmap = {}
        for ann in jsonmap.get("annotations"):
            ann._owner_set = annset
            annmap[ann.id] = ann
        annset._annotations = annmap
        annset._next_annid = jsonmap.get("next_annid")
        return annset




#!/usr/bin/env python
# encoding: utf-8

"""Packaging script for the gatenlp library."""

import os
from setuptools import setup, find_packages
import re

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as f:
    readme = f.read()

def versionfromfile(*filepath):
    here = os.path.abspath(os.path.dirname(__file__))
    infile = os.path.join(here, *filepath)
    with open(infile) as fp:
        version_match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]",
                                  fp.read(), re.M)
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string in {}.".format(infile))

setup(
    name="gatenlp",
    version=versionfromfile("gatenlp/__init__.py"),
    author="Johann Petrak",
    author_email="johann.petrak@gmail.com",
    description='GATE NLP implementation in Python.',
    long_description=readme,
    long_description_content_type='text/markdown',
    setup_requires=["pytest-runner"],
    install_requires=["sortedcontainers", "typing"],
    python_requires=">=3.4",
    tests_require=['pytest'],
    platforms='any',
    license="MIT",
    keywords="",
    url="https://github.com/GateNLP/python-gatenlp",
    packages=find_packages(),
    test_suite='tests',
    classifiers=[
        # "Development Status :: 6 - Mature",
        # "Development Status :: 5 - Production/Stable",
        # "Development Status :: 4 - Beta",
        # "Development Status :: 3 - Alpha",
        "Development Status :: 2 - Pre-Alpha",
        # "Development Status :: 1 - Planning",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
      ],
)




__version__ = "0.5"
import logging
try:
    import sortedcontainers
except Exception as ex:
    raise Exception("Required package sortedcontainers cannot be imported!")
# TODO: check version of sortedcontainers (we have 2.1.0)
logging.basicConfig()
logger = logging.getLogger("gatenlp")
logger.setLevel(logging.INFO)

# this attribute globally holds the processing resource last defined
# so it can be used for interacting with the GATE python plugin
from gatenlp.gate_interaction import _pr_decorator as GateNlpPr
from gatenlp.gate_interaction import interact
from gatenlp.annotation import Annotation
from gatenlp.document import Document
from gatenlp.annotation_set import AnnotationSet
from gatenlp.exceptions import InvalidOffsetException
from gatenlp.changelog import ChangeLog

__all__ = ["GateNlpPr", "Annotation", "Document", "AnnotationSet",
           "InvalidOffsetException", "ChangeLog", "logger"]

gate_python_plugin_pr = None

